/*****************************************************************************/
/* XMangekyou - a kaleidoscope for X                                         */
/* Programed by Sakai Hiroaki.                                               */
/*                                                                           */
/* Copyright (c) 1998 Sakai Hiroaki.                                         */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* XMangekyou is a kaleidoscope. You can see many, many, many patterns,      */
/* but, care for yourself from 'trip' filling!                               */
/*                                                                           */
/* XMangekyou is a free software. You may copy or distribute the original    */
/* XMangekyou freely. But, you may not modify or distribute the original     */
/* XMangekyou without permission of the author.                              */
/*****************************************************************************/

/*****************************************************************************/
/* Include Header File.                                                      */
/*****************************************************************************/

#include"xm_type.h"

/*****************************************************************************/
/* Functions for Animation.                                                  */
/*****************************************************************************/

void AnimationPictureStart(XMangekyou * xmp)
{
  if (xmp->animation_flag == ON) {
    xmp->animation_picture_start_time = GetTime();
    if (++(xmp->animation_picture) ==
	xmp->animations[xmp->animation_type].pictures_number)
      xmp->animation_flag = OFF;
  }
}

void SetAnimation(XMangekyou * xmp)
{
  if (xmp->animations_number > 0) {
    if (xmp->animation_fix_number == NO_ANIMATION_FIX)
      xmp->animation_type = rand() % xmp->animations_number;
    else
      xmp->animation_type = xmp->animation_fix_number;
    xmp->animation_flag = ON;
    xmp->animation_picture = 0;
    xmp->animation_picture_start_time = GetTime();
  }
}

/*****************************************************************************/
/* Free Animation's Data.                                                    */
/*****************************************************************************/

void FreeAnimationData(XMangekyou * xmp)
{
  int i, j, k;

  if (xmp->animations_number > 0) {
    for (i = 0; i < xmp->animations_number; i++) {
      for (j = 0; j < xmp->animations[i].pictures_number; j++) {
	free(xmp->animations[i].pictures[j].lines);
	free(xmp->animations[i].pictures[j].circles);
      }
      free(xmp->animations[i].pictures);
    }
    free(xmp->animations);
  }
}

/*****************************************************************************/
/* Dump Animation's Data.                                                    */
/*****************************************************************************/

void DumpAnimationData(XMangekyou * xmp)
{
  int i, j, k;

  printf("animations_number = %d\n\n", xmp->animations_number);
  for (i = 0; i < xmp->animations_number; i++) {
    printf("animations[%d].pictures_number = %d\n\n", i,
	   xmp->animations[i].pictures_number);
    for (j = 0; j < xmp->animations[i].pictures_number; j++) {
      printf("animations[%d].pictures[%d].count = %ld\n", i, j,
	     xmp->animations[i].pictures[j].count);
      printf("animations[%d].pictures[%d].lines_number = %d\n", i, j,
	     xmp->animations[i].pictures[j].lines_number);
      printf("animations[%d].pictures[%d].circles_number = %d\n", i, j,
	     xmp->animations[i].pictures[j].circles_number);
      printf("Line Data : ");
      for (k = 0; k < xmp->animations[i].pictures[j].lines_number; k++)
	printf("%d %d %d %d   ",
	       (int)(xmp->animations[i].pictures[j].lines[k].point_1.x),
	       (int)(xmp->animations[i].pictures[j].lines[k].point_1.y),
	       (int)(xmp->animations[i].pictures[j].lines[k].point_2.x),
	       (int)(xmp->animations[i].pictures[j].lines[k].point_2.y));
      printf("\nCircle Data : ");
      for (k = 0; k < xmp->animations[i].pictures[j].circles_number; k++)
	printf("%d %d   ",
	       (int)(xmp->animations[i].pictures[j].circles[k].center.x),
	       (int)(xmp->animations[i].pictures[j].circles[k].center.y));
      printf("\n\n");
    }
  }
  fflush(stdout);
}

/*****************************************************************************/
/* Make Animation's Data.                                                    */
/*****************************************************************************/

void ReadAnimationData(XMangekyou * xmp, FILE * fp)
{
  int i, j, k;
  Picture * p;

  xmp->animations_number = ReadInt(fp);
  xmp->animations =
    (Animation *)malloc(sizeof(Animation) * xmp->animations_number);
  for (i = 0; i < xmp->animations_number; i++) {
    xmp->animations[i].pictures_number = ReadInt(fp);
    xmp->animations[i].pictures =
      (Picture *)malloc(sizeof(Picture)*xmp->animations[i].pictures_number);
    for (j = 0; j < xmp->animations[i].pictures_number; j++) {
      p = &(xmp->animations[i].pictures[j]);
      p->count          = (unsigned long int)ReadInt(fp);
      p->lines_number   = ReadInt(fp);
      p->circles_number = ReadInt(fp);
      if (p->lines_number == 0) {
	p->lines_number = 1;
	p->lines = (PictureLine *)malloc(sizeof(PictureLine));
	p->lines[k].point_1.x = p->lines[k].point_1.y = 0.0;
	p->lines[k].point_2.x = p->lines[k].point_2.y = 0.0;
      } else {
	p->lines =
	  (PictureLine *)malloc(sizeof(PictureLine) * p->lines_number);
	for (k = 0; k < p->lines_number; k++) {
	  p->lines[k].point_1.x = ReadChar(fp);
	  p->lines[k].point_1.y = ReadChar(fp);
	  p->lines[k].point_2.x = ReadChar(fp);
	  p->lines[k].point_2.y = ReadChar(fp);
	}
      }
      if (p->circles_number == 0) {
	p->circles_number = 1;
	p->circles = (PictureCircle *)malloc(sizeof(PictureCircle));
	p->circles[k].center.x = p->circles[k].center.y = 0.0;
      } else {
	p->circles =
	  (PictureCircle *)malloc(sizeof(PictureCircle) * p->circles_number);
	for (k = 0; k < p->circles_number; k++) {
	  p->circles[k].center.x = ReadChar(fp);
	  p->circles[k].center.y = ReadChar(fp);
	}
      }
    }
  }
}

void MakeAnimationData(XMangekyou * xmp)
{
  FILE * fp;
  char filename[512];

  strcpy(filename, XMANGEKYOU_LIB);
  strcat(filename, xmp->animation_data_filename);
  if ((fp = fopen(xmp->animation_data_filename, "rt")) != NULL) {
    ReadAnimationData(xmp, fp);
    fclose(fp);
  } else if ((fp = fopen(filename, "rt")) != NULL) {
    ReadAnimationData(xmp, fp);
    fclose(fp);
  } else xmp->animations_number = 0;
}

/*****************************************************************************/
/* Moving Pieces in Animation.                                               */
/*****************************************************************************/

void AnimationPoint(XMangekyou * xmp, CharPoint * p1, DoublePoint * p2)
{
  double move_r, move_speed;
  DoublePoint move;

  move.x = 0.01 * xmp->size * (double)((*p1).x) - (*p2).x;
  move.y = 0.01 * xmp->size * (double)((*p1).y) - (*p2).y;
  move_r = sqrt(move.x * move.x + move.y * move.y);
  if (ANIMATION_SPEED * xmp->size < move_r)
    move_speed = ANIMATION_SPEED * xmp->size / move_r;
  else move_speed = 1.0;
  (*p2).x += move.x * move_speed;
  (*p2).y += move.y * move_speed;
}

void AnimationLines(XMangekyou * xmp)
{
  int i, a;
  Picture * p;

  p = &(xmp->animations[xmp->animation_type].pictures[xmp->animation_picture]);
  for (i = 0; i < xmp->lines_number; i++) {
    a = i % p->lines_number;
    AnimationPoint(xmp, &(p->lines[a].point_1), &(xmp->lines[i].point_1));
    AnimationPoint(xmp, &(p->lines[a].point_2), &(xmp->lines[i].point_2));
  }
}

void AnimationCircles(XMangekyou * xmp)
{
  int i, a;
  Picture * p;

  p = &(xmp->animations[xmp->animation_type].pictures[xmp->animation_picture]);
  for (i = 0; i < xmp->circles_number; i++) {
    a = i % p->circles_number;
    AnimationPoint(xmp, &(p->circles[a].center), &(xmp->circles[i].center));
  }
}

void AnimationPieces(XMangekyou * xmp)
{
  AnimationLines(xmp);
  AnimationCircles(xmp);
}

/*****************************************************************************/
/* Check Animation.                                                          */
/*****************************************************************************/

void CheckAnimation(XMangekyou * xmp)
{
  if (xmp->animation_flag == ON) {
    if (GetTime() - xmp->animation_picture_start_time >
	xmp->animations[xmp->animation_type].pictures[xmp->animation_picture
						      ].count)
      AnimationPictureStart(xmp);
  }
  else if ((GetMinute() % 30) == 0) SetAnimation(xmp);
}

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
