/*****************************************************************************/
/* XMangekyou - a kaleidoscope for X                                         */
/* Programed by Sakai Hiroaki.                                               */
/*                                                                           */
/* Copyright (c) 1998 Sakai Hiroaki.                                         */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* XMangekyou is a kaleidoscope. You can see many, many, many patterns,      */
/* but, care for yourself from 'trip' filling!                               */
/*                                                                           */
/* XMangekyou is a free software. You may copy or distribute the original    */
/* XMangekyou freely. But, you may not modify or distribute the original     */
/* XMangekyou without permission of the author.                              */
/*****************************************************************************/

/*****************************************************************************/
/* Include Header File.                                                      */
/*****************************************************************************/

#include"xm_type.h"

/*****************************************************************************/
/* Set Default Values.                                                       */
/*****************************************************************************/

void SetDefaultValues(XMangekyou * xmp)
{
  xmp->degree = 0;
  xmp->lines = (Line *)NULL;
  xmp->circles = (Circle *)NULL;
  xmp->source_pixmaps = (SourcePixmap *)NULL;
  xmp->copy_number = -1;
  xmp->copy_points = (DoublePoint *)NULL;
  xmp->copy_points_rotated = (DoublePoint *)NULL;
  xmp->animation_data_filename = DEFAULT_ANIMATION_DATA_FILENAME;
  xmp->animation_flag = OFF;
  xmp->animation_fix_number = NO_ANIMATION_FIX;

  xmp->width = DEFAULT_WIDTH;
  xmp->height = DEFAULT_HEIGHT;
  xmp->root_flag = DEFAULT_ROOT_FLAG;
  xmp->lines_number = DEFAULT_LINES_NUMBER;
  xmp->circles_number = DEFAULT_CIRCLES_NUMBER;
  xmp->line_thickness = DEFAULT_LINE_THICKNESS;
  xmp->circle_size = DEFAULT_CIRCLE_SIZE;
  xmp->size = DEFAULT_SIZE;
  xmp->speed = DEFAULT_SPEED;
  xmp->change = DEFAULT_CHANGE;
  xmp->wait = DEFAULT_WAIT;
}

/*****************************************************************************/
/* Read Command Line Parameters.                                             */
/*****************************************************************************/

void help()
{
  printf("%s", help_string);
  exit (0);
}

void ReadCommandParameter(XMangekyou * xmp, int argc, char ** argv)
{
  int i;

  for (i = 1; i < argc; i++) {
    if      (!strcmp(argv[i], "-h"     )) help();
    else if (!strcmp(argv[i], "-help"  )) help();
    else if (!strcmp(argv[i], "-root"  )) xmp->root_flag = ON;
    else if (!strcmp(argv[i], "-noroot")) xmp->root_flag = OFF;
    else if (i < argc - 1) {
      if      (!strcmp(argv[i],"-lines" )) xmp->lines_number=atoi(argv[++i]);
      else if (!strcmp(argv[i],"-circles"))xmp->circles_number=atoi(argv[++i]);
      else if (!strcmp(argv[i], "-line-thickness" ))
	xmp->line_thickness = atoi(argv[++i]);
      else if (!strcmp(argv[i], "-circle-size" ))
	xmp->circle_size = atoi(argv[++i]);
      else if (!strcmp(argv[i], "-size" )) xmp->size = atoi(argv[++i]);
      else if (!strcmp(argv[i], "-speed")) xmp->speed = atoi(argv[++i]);
      else if (!strcmp(argv[i], "-change" )) xmp->change = atoi(argv[++i]);
      else if (!strcmp(argv[i], "-wait" )) xmp->wait = atoi(argv[++i]);
      else if (!strcmp(argv[i], "-animation-fix" ))
	xmp->animation_fix_number = atoi(argv[++i]);
      else if (!strcmp(argv[i], "-animation-file" ))
	xmp->animation_data_filename = argv[++i];
    }
  }
}

/*****************************************************************************/
/* Initializing Display.                                                     */
/*****************************************************************************/

void CreateCanvas(XMangekyou * xmp, int argc, char ** argv)
{
  Window root;
  int x, y, border, depth;
  int n;
  Arg args[10];

  n = 0;
  XtSetArg(args[n], XtNwidth , xmp->width ); n++;
  XtSetArg(args[n], XtNheight, xmp->height); n++;
  xmp->toplevel = XtAppInitialize(&(xmp->app_context), "XMangekyou", NULL, 0,
				  &argc, argv, NULL, args, n);
  xmp->display = XtDisplay(xmp->toplevel);
  xmp->screen = DefaultScreen(xmp->display);
  xmp->colormap = DefaultColormap(xmp->display, xmp->screen);
  xmp->depth = DefaultDepth(xmp->display, xmp->screen);
  xmp->root_window = XRootWindow(xmp->display, xmp->screen);
  n = 0;
  XtSetArg(args[n], XtNbackground, &(xmp->background_color)); n++;
  XtGetValues(xmp->toplevel, args, n);

  if (xmp->root_flag == ON) {
    xmp->canvas_window = xmp->root_window;
    xmp->width  = DisplayWidth(xmp->display, xmp->screen);
    xmp->height = DisplayHeight(xmp->display, xmp->screen);
  } else {
    XtRealizeWidget(xmp->toplevel);
    xmp->toplevel_window = XtWindow(xmp->toplevel);

    n = 0;
    XtSetArg(args[n], XtNwidth , &(xmp->width )); n++;
    XtSetArg(args[n], XtNheight, &(xmp->height)); n++;
    XtGetValues(xmp->toplevel, args, n);

    xmp->canvas_window =
      XCreateSimpleWindow(xmp->display, xmp->toplevel_window,
			  0, 0, xmp->width, xmp->height, 0,
			  BlackPixel(xmp->display, xmp->screen),
			  BlackPixel(xmp->display, xmp->screen));
    XMapWindow(xmp->display, xmp->canvas_window);
  }
}

void CreateGC(XMangekyou * xmp)
{
  unsigned long int black;

  black = BlackPixel(xmp->display, xmp->screen);
  xmp->gc_source_clear = XCreateGC(xmp->display, xmp->canvas_window, 0, 0);
  xmp->gc_source_hex_clear = XCreateGC(xmp->display, xmp->canvas_window, 0, 0);
  xmp->gc_canvas_clear = XCreateGC(xmp->display, xmp->canvas_window, 0, 0);
  xmp->gc_copy_source_canvas = XCreateGC(xmp->display, xmp->canvas_window,0,0);
  xmp->gc_copy_canvas_canvas = XCreateGC(xmp->display, xmp->canvas_window,0,0);
  XSetForeground(xmp->display, xmp->gc_source_clear, xmp->background_color);
  XSetBackground(xmp->display, xmp->gc_source_clear, xmp->background_color);
  XSetForeground(xmp->display, xmp->gc_source_hex_clear, black);
  XSetBackground(xmp->display, xmp->gc_source_hex_clear, black);
  XSetForeground(xmp->display, xmp->gc_canvas_clear, black);
  XSetBackground(xmp->display, xmp->gc_canvas_clear, black);
  XSetFillRule(xmp->display, xmp->gc_source_hex_clear, EvenOddRule);
  XSetFunction(xmp->display, xmp->gc_copy_source_canvas, GXor);
  XSetFunction(xmp->display, xmp->gc_copy_canvas_canvas, GXcopy);
}

void InitializeDisplay(XMangekyou * xmp, int argc, char ** argv)
{
  CreateCanvas(xmp, argc, argv);
  CreateGC(xmp);
}

/*****************************************************************************/
/* Initializing Mangekyou.                                                   */
/*****************************************************************************/

int MakeHexCopyData(XMangekyou * xmp, int r)
{
  int i, j, k, a, n;
  DoublePoint dp, dv, csp, csv, p, v;

  n = 3*r*r + 3*r + 1;
  xmp->copy_points         = (DoublePoint *)malloc(sizeof(DoublePoint) * n);
  xmp->copy_points_rotated = (DoublePoint *)malloc(sizeof(DoublePoint) * n);
  xmp->copy_points[0].x = xmp->copy_points[0].y = 0.0;
  a = 1;
  dv.x = SQRT_3;
  dv.y = 0.0;
  for (i = 1; i <= r; i++) {
    dp.x = SQRT_3 * i;
    dp.y = 0.0;
    for (j = 0; j < 6; j++) {
      CalCosSin(64 *  60*j,      &csp); CalRotate(dp, csp, &p);
      CalCosSin(64 * (60*j+120), &csv); CalRotate(dv, csv, &v);
      for (k = 0; k < i; k++) {
	xmp->copy_points[a  ].x = p.x + (v.x * k);
	xmp->copy_points[a++].y = p.y + (v.y * k);
      }
    }
  }
  return (n);
}

void InitializeDrawRoutine1(XMangekyou * xmp, int r)
{
  int i;

  xmp->copy_end = xmp->copy_max = MakeHexCopyData(xmp, r);
  xmp->copy_number = 1;
  xmp->source_pixmaps = (SourcePixmap *)malloc(sizeof(SourcePixmap) * 2);
  xmp->source_pixmaps[0].width.x = xmp->size * 2 + 3;
  xmp->source_pixmaps[0].width.y = xmp->size * 2 + 3;
  xmp->source_pixmaps[1].width.x = xmp->width;
  xmp->source_pixmaps[1].width.y = xmp->height;
  xmp->source_pixmaps[0].size = xmp->size;
  for (i = 0; i < 2; i++) {
    xmp->source_pixmaps[i].center.x = xmp->source_pixmaps[i].width.x / 2;
    xmp->source_pixmaps[i].center.y = xmp->source_pixmaps[i].width.y / 2;
    xmp->source_pixmaps[i].pixmap =
      XCreatePixmap(xmp->display, xmp->canvas_window,
		    xmp->source_pixmaps[i].width.x,
		    xmp->source_pixmaps[i].width.y, xmp->depth);
  }
#ifdef XMANGEKYOU_DEBUG
  fprintf(stderr, "Draw routine 1 is selected.\n");
#endif
}

void InitializeDrawRoutine2(XMangekyou * xmp, double canvas_size)
{
  int i, a, n;
  double tmp;

  MakeHexCopyData(xmp, 2);
  tmp = log(canvas_size / (0.5 * SQRT_3 * xmp->size)) / log(2.0);
  if (tmp < 0) n = 0; else n = (int)tmp + 1;
  xmp->copy_max = 19;
  if ((n % 2) == 1) xmp->copy_end = 7; else xmp->copy_end = 19;
  xmp->copy_number = (n + 1) / 2;
  xmp->source_pixmaps =
    (SourcePixmap *)malloc(sizeof(SourcePixmap)*(xmp->copy_number + 1));
  a = 1;
  for (i = 0; i <= xmp->copy_number; i++) {
    if (i == xmp->copy_number) {
      xmp->source_pixmaps[i].width.x = xmp->width;
      xmp->source_pixmaps[i].width.y = xmp->height;
    } else {
      xmp->source_pixmaps[i].width.x = a * xmp->size * 2 + 3;
      xmp->source_pixmaps[i].width.y = a * xmp->size * 2 + 3;
    }
    xmp->source_pixmaps[i].size = a * xmp->size;
    xmp->source_pixmaps[i].center.x = xmp->source_pixmaps[i].width.x / 2;
    xmp->source_pixmaps[i].center.y = xmp->source_pixmaps[i].width.y / 2;
    xmp->source_pixmaps[i].pixmap =
      XCreatePixmap(xmp->display, xmp->canvas_window,
		    xmp->source_pixmaps[i].width.x,
		    xmp->source_pixmaps[i].width.y,
		    xmp->depth);
    a *= 4;
  }
#ifdef XMANGEKYOU_DEBUG
  fprintf(stderr, "Draw routine 2 is selected.\n");
#endif
}

void InitializeMangekyou(XMangekyou * xmp)
{
  int r;
  double canvas_size;

  canvas_size = 0.5 * sqrt((double)xmp->width  * xmp->width +
			   (double)xmp->height * xmp->height) + 2.0;
  r = (int)((canvas_size - 0.5 * xmp->size) / (1.5 * xmp->size) + 1.0);
  if (r < 11) InitializeDrawRoutine1(xmp, r);
  else InitializeDrawRoutine2(xmp, canvas_size);
}

/*****************************************************************************/
/* Initializing Pieces.                                                      */
/*****************************************************************************/

void InitializeLines(XMangekyou * xmp)
{
  int i;

  xmp->lines = (Line *)malloc(sizeof(Line) * xmp->lines_number);
  for (i = 0; i < xmp->lines_number; i++) {
    xmp->lines[i].point_1.x = (double)(rand() % xmp->size);
    xmp->lines[i].point_1.y = (double)(rand() % xmp->size) - (xmp->size / 2);
    xmp->lines[i].point_2.x = (double)(rand() % xmp->size);
    xmp->lines[i].point_2.y = (double)(rand() % xmp->size) - (xmp->size / 2);
    xmp->lines[i].gc = XCreateGC(xmp->display, xmp->canvas_window, 0, 0);
    XSetForeground(xmp->display, xmp->lines[i].gc,
		   GetColor(xmp, color_name[rand() % COLORS_NUMBER]));
    XSetLineAttributes(xmp->display, xmp->lines[i].gc,
		       (rand() % xmp->line_thickness),
		       LineSolid, CapButt ,JoinMiter);
  }
}

void InitializeCircles(XMangekyou * xmp)
{
  int i;

  xmp->circles = (Circle *)malloc(sizeof(Circle) * xmp->circles_number);
  for (i = 0; i < xmp->circles_number; i++) {
    xmp->circles[i].r = (rand() % xmp->circle_size);
    xmp->circles[i].center.x = (double)(rand() % xmp->size);
    xmp->circles[i].center.y = (double)(rand() % xmp->size) - (xmp->size / 2);
    xmp->circles[i].gc = XCreateGC(xmp->display, xmp->canvas_window, 0, 0);
    XSetForeground(xmp->display, xmp->circles[i].gc,
		   GetColor(xmp, color_name[rand() % COLORS_NUMBER]));
    XSetArcMode(xmp->display, xmp->circles[i].gc, ArcPieSlice);
  }
}

void InitializePieces(XMangekyou * xmp)
{
  srand((unsigned)time(NULL));
  InitializeLines(xmp);
  InitializeCircles(xmp);
}

/*****************************************************************************/
/* Addition of EventHandler                                                  */
/*****************************************************************************/

void PushKey(Widget w, XtPointer p, XEvent * event,
	     Boolean * dispatch)
{
  XMangekyou * xmp;

  xmp = (XMangekyou *)p;
  if (event->type == KeyPress) {
    if (XKeycodeToKeysym(event->xkey.display, event->xkey.keycode, 0) == XK_q)
      Quit(xmp);
    if (XKeycodeToKeysym(event->xkey.display, event->xkey.keycode, 0) == XK_a)
      SetAnimation(xmp);
    if (XKeycodeToKeysym(event->xkey.display, event->xkey.keycode, 0) == XK_p)
      AnimationPictureStart(xmp);
    if (XKeycodeToKeysym(event->xkey.display, event->xkey.keycode, 0) == XK_d)
      DumpAnimationData(xmp);
  }
#ifdef XMANGEKYOU_DEBUG
  fprintf(stderr, "Key Press.\n");
#endif
}

void ExposeCanvas(Widget w, XtPointer p, XEvent * event,
		  Boolean * dispatch)
{
  XMangekyou * xmp;

  xmp = (XMangekyou *)p;
  if (event->type == Expose) {
    XCopyArea(xmp->display, xmp->source_pixmaps[xmp->copy_number].pixmap,
	      xmp->canvas_window, xmp->gc_copy_canvas_canvas,
	      event->xexpose.x, event->xexpose.y,
	      event->xexpose.width, event->xexpose.height,
	      event->xexpose.x, event->xexpose.y);
  }
#ifdef XMANGEKYOU_DEBUG
  fprintf(stderr, "Expose.\n");
#endif
}

void ResizeCanvas(XMangekyou * xmp)
{
  int i;
  int n = 0;
  Arg args[10];

  XtSetArg(args[n], XtNwidth, &(xmp->width)); n++;
  XtSetArg(args[n], XtNheight, &(xmp->height)); n++;
  XtGetValues(xmp->toplevel, args, n);
  XResizeWindow(xmp->display, xmp->canvas_window, xmp->width, xmp->height);
  for (i = 0; i <= xmp->copy_number; i++)
    XFreePixmap(xmp->display, xmp->source_pixmaps[i].pixmap);
  free(xmp->source_pixmaps);
  free(xmp->copy_points);
  free(xmp->copy_points_rotated);
  InitializeMangekyou(xmp);
#ifdef XMANGEKYOU_DEBUG
  fprintf(stderr, "Resize.\n");
#endif
}

void CanvasEvent(Widget w, XtPointer p, XEvent * event, Boolean * dispatch)
{
  XMangekyou * xmp;

  xmp = (XMangekyou *)p;
  if (event->type == ConfigureNotify) ResizeCanvas(xmp);
  else if (event->type == ClientMessage) {
    if ( (event->xclient.message_type == xmp->atom_1) &&
	 (event->xclient.data.l[0] == xmp->atom_2) ) Quit(xmp);
  }
#ifdef XMANGEKYOU_DEBUG
  fprintf(stderr, "Structure Notify Event.\n");
#endif
}

void AddEvent(XMangekyou * xmp)
{
  XtAddEventHandler(xmp->toplevel, ExposureMask, False, ExposeCanvas, xmp);
  if (xmp->root_flag == OFF) {
    XtAddEventHandler(xmp->toplevel, KeyPressMask, False, PushKey, xmp);
    XtAddEventHandler(xmp->toplevel, StructureNotifyMask, True,
		      CanvasEvent, xmp);
  }
}

/*****************************************************************************/
/* Setting Icon.                                                             */
/*****************************************************************************/

void SetIcon(XMangekyou * xmp)
{
  XIconSize * is;
  int i, j, n;
  int width_match = OFF;
  int height_match = OFF;
  XWMHints wmh;

  xmp->icon_pixmap =
    MakePixmapFromData(xmp->display, xmp->root_window, XMANGEKYOU_ICON,
		       &(xmp->icon_width), &(xmp->icon_height));

  wmh.flags = IconPixmapHint;
  wmh.icon_pixmap = xmp->icon_pixmap;
  XSetWMHints(xmp->display, xmp->toplevel_window, &wmh);
}

/*****************************************************************************/
/* Setting Window Attributes.                                                */
/*****************************************************************************/

void SetWindowAttributes(XMangekyou * xmp)
{
  XWMHints wmhints;

  if (xmp->root_flag == OFF) {
    XtSetSensitive(xmp->toplevel, True);
    wmhints.input = True;
    XSetWMHints(xmp->display, xmp->toplevel_window, &wmhints);
    xmp->atom_1 = XInternAtom(xmp->display, "WM_PROTOCOLS", False);
    xmp->atom_2 = XInternAtom(xmp->display, "WM_DELETE_WINDOW", False);
    XSetWMProtocols(xmp->display, xmp->toplevel_window, &(xmp->atom_2), 1);
    XStoreName(xmp->display, xmp->toplevel_window, "XMangekyou");
    XSetIconName(xmp->display, xmp->toplevel_window, "XMangekyou");
    SetIcon(xmp);
  }
}

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
