/*****************************************************************************/
/* XMangekyou - a kaleidoscope for X                                         */
/* Programed by Sakai Hiroaki.                                               */
/*                                                                           */
/* Copyright (c) 1998 Sakai Hiroaki.                                         */
/* All Rights Reserved.                                                      */
/*                                                                           */
/* XMangekyou is a kaleidoscope. You can see many, many, many patterns,      */
/* but, care for yourself from 'trip' filling!                               */
/*                                                                           */
/* XMangekyou is a free software. You may copy or distribute the original    */
/* XMangekyou freely. But, you may not modify or distribute the original     */
/* XMangekyou without permission of the author.                              */
/*****************************************************************************/

/*****************************************************************************/
/* Include Header File.                                                      */
/*****************************************************************************/

#include"xm_type.h"

/*****************************************************************************/
/* Moving Pieces.                                                            */
/*****************************************************************************/

void MovePieces(XMangekyou * xmp)
{
  int i;
  DoublePoint dominant, cs, move;
  dominant.x = 1.0;
  dominant.y = 0.0;
  CalCosSin(- xmp->degree, &cs);
  CalRotate(dominant, cs, &move);
  move.x *= 0.000001 * xmp->change;
  move.y *= 0.000001 * xmp->change;
  for (i = 0; i < xmp->lines_number; i++) {
    xmp->lines[i].point_1.x += move.x * (rand()%1000);
    xmp->lines[i].point_1.y += move.y * (rand()%1000);
    xmp->lines[i].point_2.x += move.x * (rand()%1000);
    xmp->lines[i].point_2.y += move.y * (rand()%1000);
  }
  for (i = 0; i < xmp->circles_number; i++) {
    xmp->circles[i].center.x += move.x * (rand()%1000);
    xmp->circles[i].center.y += move.y * (rand()%1000);
  }
}

/*****************************************************************************/
/* Drawing Mangekyou.                                                        */
/*****************************************************************************/

SixPoints CalHexPoints(DoublePoint p1, DoublePoint value_000,
		       DoublePoint value_120, DoublePoint value_240)
{
  SixPoints p;
  DoublePoint p2;
  p2.x = - p1.x;
  p2.y =   p1.y;
  CalRotate(p1, value_000, &(p.point[0]));
  CalRotate(p2, value_000, &(p.point[1]));
  CalRotate(p1, value_120, &(p.point[2]));
  CalRotate(p1, value_240, &(p.point[3]));
  CalRotate(p2, value_120, &(p.point[4]));
  CalRotate(p2, value_240, &(p.point[5]));
  return (p);
}

void ClearPixmaps(XMangekyou * xmp)
{
  int i;

  XFillRectangle(xmp->display, xmp->source_pixmaps[0].pixmap,
		 xmp->gc_source_clear, 0, 0,
		 xmp->source_pixmaps[0].width.x,
		 xmp->source_pixmaps[0].width.y);
  for (i = 1; i <= xmp->copy_number; i++)
    XFillRectangle(xmp->display, xmp->source_pixmaps[i].pixmap,
		   xmp->gc_canvas_clear, 0, 0,
		   xmp->source_pixmaps[i].width.x,
		   xmp->source_pixmaps[i].width.y);
}

void DrawPieces(XMangekyou * xmp, DoublePoint value_000,
		DoublePoint value_120, DoublePoint value_240)
{
  int i, j, n, sx, sy;
  SixPoints points_1, points_2;

  n = xmp->lines_number;
  if (n < xmp->circles_number) n = xmp->circles_number;
  sx = xmp->source_pixmaps[0].center.x;
  sy = xmp->source_pixmaps[0].center.y;
  for (i = 0; i < n; i++) {
    if (i < xmp->lines_number) {
      points_1 = CalHexPoints(xmp->lines[i].point_1,
			      value_000, value_120, value_240);
      points_2 = CalHexPoints(xmp->lines[i].point_2,
			      value_000, value_120, value_240);
      for (j = 0; j < 6; j++)
	XDrawLine(xmp->display, xmp->source_pixmaps[0].pixmap,
		  xmp->lines[i].gc,
		  (int)points_1.point[j].x+sx, (int)points_1.point[j].y+sy,
		  (int)points_2.point[j].x+sx, (int)points_2.point[j].y+sy);
    }
    if (i < xmp->circles_number) {
      points_1 = CalHexPoints(xmp->circles[i].center,
			      value_000, value_120, value_240);
      for (j = 0; j < 6; j++)
	XFillArc(xmp->display, xmp->source_pixmaps[0].pixmap,
		 xmp->circles[i].gc,
		 (int)(points_1.point[j].x - xmp->circles[i].r * 0.5) + sx,
		 (int)(points_1.point[j].y - xmp->circles[i].r * 0.5) + sy,
		 xmp->circles[i].r, xmp->circles[i].r, 0, 360*64);
    }
  }
}

void CalCopyPoints(XMangekyou * xmp, DoublePoint d)
{
  int i;

  for (i = 0; i < xmp->copy_max; i++)
    CalRotate(xmp->copy_points[i], d, &(xmp->copy_points_rotated[i]));
}

void CalClearPoints(XMangekyou * xmp, DoublePoint d)
{
  int i;

  for (i = 0; i < CLEAR_POINTS_NUMBER; i++) {
    CalRotate(clear_points[i], d, &(xmp->clear_points_rotated[i]));
    CalRotate(clear_points_offset[i], d,
	      &(xmp->clear_points_offset_rotated[i]));
  }
}

void ClearHexPoints(XMangekyou * xmp, int a)
{
  int i;

  for (i = 0; i < CLEAR_POINTS_NUMBER; i++) {
    xmp->clear_points_rotated_int[i].x =
      (int)(xmp->clear_points_rotated[i].x * xmp->source_pixmaps[a].size +
	    xmp->clear_points_offset_rotated[i].x +
	    xmp->source_pixmaps[a].center.x);
    xmp->clear_points_rotated_int[i].y =
      (int)(xmp->clear_points_rotated[i].y * xmp->source_pixmaps[a].size +
	    xmp->clear_points_offset_rotated[i].y +
	    xmp->source_pixmaps[a].center.y);
  }
  XFillPolygon(xmp->display, xmp->source_pixmaps[a].pixmap,
	       xmp->gc_source_hex_clear,
	       xmp->clear_points_rotated_int, CLEAR_POINTS_NUMBER,
	       Complex, CoordModeOrigin);
}

void CopyPixmaps(XMangekyou * xmp, int a)
{
  int b, i, cx, cy;

  if (a == xmp->copy_number - 1) b = xmp->copy_end; else b = xmp->copy_max;
  for (i = 0; i < b; i++) {
    cx = (int)(xmp->copy_points_rotated[i].x * xmp->source_pixmaps[a].size) +
      xmp->source_pixmaps[a + 1].center.x;
    cy = (int)(xmp->copy_points_rotated[i].y * xmp->source_pixmaps[a].size) +
      xmp->source_pixmaps[a + 1].center.y;
    if ( (cx >= - xmp->source_pixmaps[a].center.x) &&
	 (cx <=   xmp->source_pixmaps[a].center.x +
	  xmp->source_pixmaps[a+1].width.x) &&
	 (cy >= - xmp->source_pixmaps[a].center.y) &&
	 (cy <=   xmp->source_pixmaps[a].center.y +
	  xmp->source_pixmaps[a+1].width.y) )
      XCopyArea(xmp->display, xmp->source_pixmaps[a].pixmap,
		xmp->source_pixmaps[a+1].pixmap,
		xmp->gc_copy_source_canvas, 0, 0,
		xmp->source_pixmaps[a].width.x, xmp->source_pixmaps[a].width.y,
		(int)(cx - xmp->source_pixmaps[a].center.x),
		(int)(cy - xmp->source_pixmaps[a].center.y));
  }
}

void DrawMangekyou(XMangekyou * xmp)
{
  int i;
  DoublePoint value_000, value_120, value_240;

  CalCosSin(         xmp->degree, &value_000);
  CalCosSin(120*64 + xmp->degree, &value_120);
  CalCosSin(240*64 + xmp->degree, &value_240);
  ClearPixmaps(xmp);
  DrawPieces(xmp, value_000, value_120, value_240);
  CalCopyPoints(xmp, value_000);
  CalClearPoints(xmp, value_000);

  for (i = 0; i < xmp->copy_number; i++) {
    ClearHexPoints(xmp, i);
    CopyPixmaps(xmp, i);
  }

  XCopyArea(xmp->display, xmp->source_pixmaps[xmp->copy_number].pixmap,
	    xmp->canvas_window, xmp->gc_copy_canvas_canvas, 0, 0,
	    xmp->width, xmp->height, 0, 0);
  XFlush(xmp->display);
}

/*****************************************************************************/
/* Main Loop.                                                                */
/*****************************************************************************/

Boolean MainMangekyou(caddr_t p)
{
  XMangekyou * xmp;

  xmp = (XMangekyou *)p;
  DrawMangekyou(xmp);
  if (xmp->animation_flag == ON) AnimationPieces(xmp); else MovePieces(xmp);
  xmp->degree += xmp->speed;
  if ( xmp->degree >  360*64 ) xmp->degree -= 360*64;
  if ( xmp->degree < -360*64 ) xmp->degree += 360*64;
  CheckAnimation(xmp);
  usleep(xmp->wait);
  return (False);
}

main(int argc, char ** argv)
{
  int n;
  XMangekyou xmangekyou;

  SetDefaultValues(&xmangekyou);
  ReadCommandParameter(&xmangekyou, argc, argv);
  InitializeDisplay(&xmangekyou, argc, argv);
  InitializeMangekyou(&xmangekyou);
  InitializePieces(&xmangekyou);
  MakeAnimationData(&xmangekyou);
  AddEvent(&xmangekyou);
  SetWindowAttributes(&xmangekyou);
  XtAppAddWorkProc(xmangekyou.app_context, (XtWorkProc)MainMangekyou,
		   &xmangekyou);
  XtAppMainLoop(xmangekyou.app_context);
}

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
